from decimal import Decimal
from typing import Any, Dict

RESPONSE = "Ds_Response"
DATE = "Ds_Date"
HOUR = "Ds_Hour"
MERCHANT_CODE = "Ds_MerchantCode"
TERMINAL = "Ds_Terminal"
TRANSACTION_TYPE = "Ds_TransactionType"
ORDER = "Ds_Order"
CURRENCY = "Ds_Currency"
AMOUNT = "Ds_Amount"
MERCHANT_DATA = "Ds_MerchantData"
CONSUMER_LANGUAGE = "Ds_ConsumerLanguage"
CARD_COUNTRY = "Ds_Card_Country"
CARD_TYPE = "Ds_Card_Type"
CARD_BRAND = "Ds_Card_Brand"
MERCHANT_COF_TXNID = "Ds_Merchant_Cof_Txnid"
SECURE_PAYMENT = "Ds_SecurePayment"
AUTHORIZATION_CODE = "Ds_AuthorisationCode"
ERROR_CODE = "Ds_ErrorCode"
PROCESSED_PAY_METHOD = "Ds_ProcessedPayMethod"

RESPONSE_MAP = {
    "0000": "Transacción autorizada para pagos y preautorizaciones",
    "900": "Transacción autorizada para devoluciones y confirmaciones",
    "400": "Transacción autorizada para anulaciones",
    "101": "Tarjeta caducada",
    "102": "Tarjeta en excepción transitoria o bajo sospecha de fraude",
    "106": "Intentos de PIN excedidos",
    "125": "Tarjeta no efectiva",
    "129": "Código de seguridad (CVV2/CVC2) incorrecto",
    "180": "Tarjeta ajena al servicio",
    "184": "Error en la autenticación del titular",
    "190": "Denegación del emisor sin especificar motivo",
    "191": "Fecha de caducidad errónea",
    "202": "Tarjeta en excepción transitoria o bajo sospecha de fraude con retirada de tarjeta",
    "904": "Comercio no registrado en FUC",
    "909": "Error de sistema",
    "912": "Emisor no disponible",
    "913": "Pedido repetido",
    "944": "Sesión Incorrecta",
    "950": "Operación de devolución no permitida",
    "9912": "Emisor no disponible",
    "9064": "Número de posiciones de la tarjeta incorrecto",
    "9078": "Tipo de operación no permitida para esa tarjeta",
    "9093": "Tarjeta no existente",
    "9094": "Rechazo servidores internacionales",
    "9104": "Comercio con “titular seguro” y titular sin clave de compra segura",
    "9218": "El comercio no permite op. seguras por entrada /operaciones",
    "9253": "Tarjeta no cumple el check-digit",
    "9256": "El comercio no puede realizar preautorizaciones",
    "9257": "Esta tarjeta no permite operativa de preautorizaciones",
    "9261": "Operación detenida por superar el control de restricciones en la entrada al SIS",
    "9913": "Error en la confirmación que el comercio envía al TPV Virtual",
    "9914": "Confirmación “KO” del comercio",
    "9915": "A petición del usuario se ha cancelado el pago",
    "9928": "Anulación de autorización en diferido realizada por el SIS (proceso batch)",
    "9929": "Anulación de autorización en diferido realizada por el comercio",
    "9997": "Se está procesando otra transacción en SIS con la misma tarjeta",
    "9998": "Operación en proceso de solicitud de datos de tarjeta",
    "9999": "Operación que ha sido redirigida al emisor a autenticar",
}

MERCHANT_PARAMETERS_MAP = {
    "response": RESPONSE,
    "date": DATE,
    "hour": HOUR,
    "merchant_code": MERCHANT_CODE,
    "terminal": TERMINAL,
    "transaction_type": TRANSACTION_TYPE,
    "order": ORDER,
    "currency": CURRENCY,
    "amount": AMOUNT,
    "merchant_data": MERCHANT_DATA,
    "consumer_language": CONSUMER_LANGUAGE,
    "card_country": CARD_COUNTRY,
    "card_type": CARD_TYPE,
    "card_brand": CARD_BRAND,
    "merchant_cof_txnid": MERCHANT_COF_TXNID,
    "secure_payment": SECURE_PAYMENT,
    "authorization_code": AUTHORIZATION_CODE,
    "error_code": ERROR_CODE,
    "processed_pay_method": PROCESSED_PAY_METHOD,
}


class Response:
    """
    Defines a response
    """

    _parameters = {}

    def __init__(self, parameters: Dict[str, Any]):
        MERCHANT_PARAMETERS_MAP_REVERSE = {
            value: key for key, value in MERCHANT_PARAMETERS_MAP.items()
        }
        for key, value in parameters.items():
            clean = getattr(
                self, "clean_%s" % MERCHANT_PARAMETERS_MAP_REVERSE[key], None
            )
            self._parameters[MERCHANT_PARAMETERS_MAP_REVERSE[key]] = (
                clean(value) if clean else value
            )

    def __getattr__(self, item: str) -> Any:
        if item in MERCHANT_PARAMETERS_MAP:
            return self._parameters[item]

    def __setattr__(self, key: str, value: Any):
        if key in MERCHANT_PARAMETERS_MAP:
            clean = getattr(self, "clean_%s" % MERCHANT_PARAMETERS_MAP[key], None)
            self._parameters[MERCHANT_PARAMETERS_MAP[key]] = (
                clean(value) if clean else value
            )

    @property
    def is_authorized(self):
        return (0 <= self.code <= 99) or self.code == 900 or self.code == 400

    @property
    def is_paid(self):
        return 0 <= self.code <= 99

    @property
    def is_refunded(self):
        return self.code == 900

    @property
    def is_canceled(self):
        return self.code == 400

    @property
    def code(self):
        return int(self.response)

    @property
    def message(self):
        return RESPONSE_MAP["0000"] if self.is_paid else RESPONSE_MAP[self.response]

    @staticmethod
    def clean_amount(value):
        return Decimal("%s.%s" % (str(value)[:-2], str(value)[-2:]))



from redsys.response import Response


class TestResponse:
    def test_create_response(self):
        parameters = {
            "Ds_Response": "90",
            "Ds_MerchantCode": "1056",
            "Ds_Terminal": "1",
            "Ds_TransactionType": "1",
            "Ds_Order": "000000001",
            "Ds_Amount": "10054",
            "Ds_MerchantData": "test merchant data",
            "Ds_Currency": 978,
        }
        response = Response(parameters)
        assert response.code == 90
        assert (
            response.message == "Transacción autorizada para pagos y preautorizaciones"
        )
        assert len(response._parameters.keys()) == 8
        assert response.is_authorized is True
        assert response.is_paid is True
        assert response.is_canceled is False
        assert response.is_refunded is False



import re
from decimal import Decimal
from typing import Dict

from redsys.constants import CURRENCIES, LANGUAGES, TRANSACTIONS

# General parameters
MERCHANT_CODE = "Ds_Merchant_MerchantCode"
TERMINAL = "Ds_Merchant_Terminal"
TRANSACTION_TYPE = "Ds_Merchant_TransactionType"
ORDER = "Ds_Merchant_Order"
CURRENCY = "Ds_Merchant_Currency"
AMOUNT = "Ds_Merchant_Amount"

# Recurring transaction parameters
SUM_TOTAL = "Ds_Merchant_SumTotal"
DATE_FREQUENCY = "Ds_Merchant_DateFrecuency"
CHARGE_EXPIRY_DATE = "Ds_Merchant_ChargeExpiryDate"
TRANSACTION_DATE = "Ds_Merchant_TransactionDate"
AUTHORIZATION_CODE = "Ds_Merchant_AuthorisationCode"

# Optional parameters
MERCHANT_DATA = "Ds_Merchant_MerchantData"

# Redirect client paramenters
MERCHANT_NAME = "Ds_Merchant_MerchantName"
PRODUCT_DESCRIPTION = "Ds_Merchant_ProductDescription"
TITULAR = "Ds_Merchant_Titular"
MERCHANT_URL = "Ds_Merchant_MerchantURL"
URL_OK = "Ds_Merchant_UrlOK"
URL_KO = "Ds_Merchant_UrlKO"
CONSUMER_LANGUAGE = "Ds_Merchant_ConsumerLanguage"

# Credit card data parameters
PAN = "Ds_Merchant_Pan"
EXPIRY_DATE = "Ds_Merchant_ExpiryDate"
CVV2 = "Ds_Merchant_Cvv2"

MERCHANT_PARAMETERS_MAP = {
    "merchant_code": MERCHANT_CODE,
    "terminal": TERMINAL,
    "transaction_type": TRANSACTION_TYPE,
    "order": ORDER,
    "currency": CURRENCY,
    "amount": AMOUNT,
    "sum_total": SUM_TOTAL,
    "date_frequency": DATE_FREQUENCY,
    "charge_expiry_date": CHARGE_EXPIRY_DATE,
    "transaction_date": TRANSACTION_DATE,
    "authorization_code": AUTHORIZATION_CODE,
    "merchant_data": MERCHANT_DATA,
    "merchant_name": MERCHANT_NAME,
    "product_description": PRODUCT_DESCRIPTION,
    "titular": TITULAR,
    "merchant_url": MERCHANT_URL,
    "url_ok": URL_OK,
    "url_ko": URL_KO,
    "consumer_language": CONSUMER_LANGUAGE,
    "pan": PAN,
    "expiry_date": EXPIRY_DATE,
    "cvv2": CVV2,
}


class Request:
    """
    Defines an atomic request with all the required parameters and sanitize
    their values according to the platform specifications
    """

    _parameters = {}

    def __init__(self, parameters: Dict[str, str]) -> None:
        for key in parameters.keys():
            if key in MERCHANT_PARAMETERS_MAP:
                check = getattr(self, "check_%s" % key, None)
                if check:
                    check(parameters[key])
                self._parameters[key] = parameters[key]
            else:
                raise ValueError(f"Unknown parameter {key}")

    def __getattr__(self, item: str) -> str:
        if item in MERCHANT_PARAMETERS_MAP:
            return self._parameters[item]

    def __setattr__(self, key, value):
        if key in MERCHANT_PARAMETERS_MAP:
            if check := getattr(self, "check_%s" % key, None):
                check(value)
            self._parameters[key] = value

    def prepare_parameters(self) -> Dict[str, str]:
        parameters = {}
        for key, value in self._parameters.items():
            prepare = getattr(self, "prepare_%s" % key, None)
            parameters[MERCHANT_PARAMETERS_MAP[key]] = (
                prepare(value) if prepare else value
            )
        return parameters

    @staticmethod
    def prepare_amount(value):
        return int(value * 100)

    @staticmethod
    def prepare_sum_total(value):
        return int(value * 100)

    @staticmethod
    def check_order(value):
        if not re.match(r"[0-9]{4}[a-zA-Z0-9]{5}$", value):
            raise ValueError("order format is not valid.")

    @staticmethod
    def check_transaction_type(value):
        if value not in TRANSACTIONS:
            raise ValueError("transaction_type is not valid.")

    @staticmethod
    def check_currency(value):
        if value not in CURRENCIES:
            raise ValueError("currency is not valid.")

    @staticmethod
    def check_amount(value):
        if not isinstance(value, Decimal):
            raise TypeError("amount must be defined as decimal.Decimal.")

    @staticmethod
    def check_sum_total(value):
        if type(value) is not Decimal:
            raise TypeError("sum_total must be defined as decimal.Decimal.")

    @staticmethod
    def check_merchant_data(value):
        if len(value) > 1024:
            raise ValueError("merchant_data is bigger than 1024 characters.")

    @staticmethod
    def check_merchant_url(value):
        if len(value) > 250:
            raise ValueError("merchant_url is bigger than 250 characters.")

    @staticmethod
    def check_url_ok(value):
        if len(value) > 250:
            raise ValueError("url_ok is bigger than 250 characters.")

    @staticmethod
    def check_url_ko(value):
        if len(value) > 250:
            raise ValueError("url_ko is bigger than 250 characters.")

    @staticmethod
    def check_consumer_language(value):
        if value not in LANGUAGES:
            raise ValueError("consumer_language is not valid.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['redsys', 'redsys.tests']

package_data = \
{'': ['*']}

install_requires = \
['pycryptodome>=3.9.9,<4.0.0']

setup_kwargs = {
    'name': 'python-redsys',
    'version': '1.0.0',
    'description': 'A simple, clean and less dependant client to handle payments through Redsys.',
    'long_description': '\n[![PyPI version](https://badge.fury.io/py/python-redsys.svg)](https://badge.fury.io/py/python-redsys)\n\n# Welcome to python-redsys!\n\nA simple, clean and less dependant client to handle payments through the\nRedsys platform using one of the available methods: _redirect connection_ or (secure method).\n\nThe purpose of this library is to provide a normalized interface between\nRedsys and other Python applications.\n\n**About RedirectClient**\n\nAlthough _redirect connection_ depends on a webserver to resolve the\ncommunication step, the RedirectClient provided in this library does not\nassume any kind of procedure to resolve that step; it merely prepares\nthe necessary parameters to make a request and handles the corresponding\nresponse parameters. That\'s what less dependant means.\n\n## Example using _redirect connection_\n\n### 0. Install python-redsys\n\nYou can add python-redsys to your project with pip:\n> pip install python-redsys\n\nOr with poetry:\n> poetry add python-redsys\n\n### 1. Instantiate the redirect client\n\n```python\nfrom decimal import Decimal as D, ROUND_HALF_UP\nfrom redsys.constants import EUR, STANDARD_PAYMENT\nfrom redsys.client import RedirectClient\n\nsecret_key = "123456789abcdef"\nclient = RedirectClient(secret_key)\n```\n\n### 2. Set up the request parameters\n\n```python\nparameters = {\n  "merchant_code": "100000001",\n  "terminal": "1",\n  "transaction_type": STANDARD_PAYMENT,\n  "currency": EUR,\n  "order": "000000001",\n  "amount": D("10.56489").quantize(D(".01"), ROUND_HALF_UP),\n  "merchant_data": "test merchant data",\n  "merchant_name": "Example Commerce",\n  "titular": "Example Ltd.",\n  "product_description": "Products of Example Commerce",\n  "merchant_url": "https://example.com/redsys/response",\n}\n```\n\n### 3. Prepare the request\n\nThis method returns a dict with the necessary post parameters that are\nneeded during the communication step.\n\n```python\nargs = client.prepare_request(parameters)\n```\n\n### 4. Communication step\n\nRedirect the _user-agent_ to the corresponding Redsys\' endpoint using\nthe post parameters given in the previous step.\n\nAfter the payment process is finish, Redsys will respond making a\nrequest to the `merchant_url` defined in step 2.\n\n### 5. Create and check the response\n\nCreate the response object using the received parameters from Redsys.\nThe method `create_response()` throws a `ValueError` in case the\nreceived signature is not equal to the calculated one using the\ngiven `merchant_parameters`. This normally means that the response **is\nnot coming from Redsys** or that it **has been compromised**.\n\n```python\nsignature = "YqFenHc2HpB273l8c995...."\nmerchant_parameters = "AndvIh66VZdkC5TG3nYL5j4XfCnFFbo3VkOu9TAeTs58fxddgc..."\nresponse = client.create_response(signature, merchant_parameters)\nif response.is_paid():\n    # Do the corresponding actions after a successful payment\nelse:\n    # Do the corresponding actions after a failed payment\n    raise Exception(response.code, response.message)\n```\n\n**Methods for checking the response:**\n\nAccording to the Redsys documentation:\n\n- `response.is_paid`: Returns `True` if the response code is\n  between 0 and 99 (both included).\n- `response.is_canceled`: Returns `True` if the response code\n  is 400.\n- `response.is_refunded`: Returns `True` if the response code\n  is 900.\n- `response.is_authorized`: Returns `True` if the response is\n  **paid**, **refunded** or **canceled**.\n\nAlso, you can directly access the code or the message defined in Redsys\ndocumentation using `response.code` or `response.message`.\n\n## Contributions\n\nPlease, feel free to send any contribution that maintains the _less\ndependant_ philosophy.\n',
    'author': 'Andrés Reverón Molina',
    'author_email': 'andres@reveronmolina.me',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/systemallica/python-redsys',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# Currencies
""" Euro """
EUR = 978
""" United States dollar """
USD = 840
""" Pound sterling """
GBP = 826
""" Japanese yen """
JPY = 392
""" Argentine peso """
ARS = 32
""" Canadian dollar """
CAD = 124
""" Chilean Peso """
CLP = 152
""" Colombian peso """
COP = 170
""" Indian rupee """
INR = 356
""" Mexican peso """
MXN = 484
""" Peruvian Sol """
PEN = 604
""" Swiss franc """
CHF = 756
""" Brazilian real """
BRL = 986
""" Venezuelan bolivar """
VEF = 937
""" Turkish lira """
TRY = 949

CURRENCIES = [EUR, USD, GBP, JPY, ARS, CAD, CLP, COP, INR, MXN, PEN, CHF, BRL, VEF, TRY]

# Languages
CUSTOMER = "000"
SPANISH = "001"
ENGLISH = "002"
CATALAN = "003"
FRENCH = "004"
GERMAN = "005"
DUTCH = "006"
ITALIAN = "007"
SWEDISH = "008"
PORTUGUESE = "009"
VALENCIAN = "010"
POLISH = "011"
GALICIAN = "012"
EUSKERA = "013"

LANGUAGES = [
    CUSTOMER,
    SPANISH,
    ENGLISH,
    CATALAN,
    FRENCH,
    GERMAN,
    DUTCH,
    ITALIAN,
    SWEDISH,
    PORTUGUESE,
    VALENCIAN,
    POLISH,
    GALICIAN,
    EUSKERA,
]

# Transaction types
""" Standard payment """
STANDARD_PAYMENT = "0"
""" Pre-authorization """
PREAUTHORIZATION = "1"
""" Confirmation of pre-authorization """
PREAUTHORIZATION_CONFIRMATION = "2"
""" Partial or total refund """
REFUND = "3"
""" Recurring transaction """
RECURRING_TRANSACTION = "5"
""" Successive transaction """
SUCCESSIVE_TRANSACTION = "6"
""" Authentication """
AUTHENTICATION = "7"
""" Confirmation of authentication """
AUTHENTICATION_CONFIRMATION = "8"
""" Cancellation of pre-authorization """
PREAUTHORIZATION_CANCELATION = "9"
""" Deferred pre-authorization """
DEFERRED_PREAUTHORIZATION = "O"
""" Confirmation of deferred pre-authorization """
DEFERRED_PREAUTHORIZATION_CONFIRMATION = "P"
""" Cancelation of deferred pre-authorization """
DEFERRED_PREAUTHORIZATION_CANCELATION = "Q"
""" Recurring deferred pre-authorization """
RECURRING_DEFERRED_PREAUTHORIZATION = "R"
""" Confirmation of recurring deferred pre-authorization and successive transaction """
SUCCESSIVE_RECURRING_TRANSACTION = "S"

TRANSACTIONS = [
    STANDARD_PAYMENT,
    PREAUTHORIZATION,
    PREAUTHORIZATION_CONFIRMATION,
    REFUND,
    RECURRING_TRANSACTION,
    SUCCESSIVE_TRANSACTION,
    AUTHENTICATION,
    AUTHENTICATION_CONFIRMATION,
    PREAUTHORIZATION_CANCELATION,
    DEFERRED_PREAUTHORIZATION,
    DEFERRED_PREAUTHORIZATION_CONFIRMATION,
    DEFERRED_PREAUTHORIZATION_CANCELATION,
    RECURRING_DEFERRED_PREAUTHORIZATION,
    SUCCESSIVE_RECURRING_TRANSACTION,
]



from decimal import ROUND_HALF_UP
from decimal import Decimal as D

import pytest

from redsys.client import RedirectClient
from redsys.constants import EUR, STANDARD_PAYMENT
from redsys.request import Request


class TestRedirectClient:
    @pytest.fixture(autouse=True)
    def set_up(self):
        secret_key = "sq7HjrUOBfKmC576ILgskD5srU870gJ7"
        self.client = RedirectClient(secret_key)
        self.parameters = {
            "merchant_code": "100000001",
            "terminal": "1",
            "transaction_type": STANDARD_PAYMENT,
            "currency": EUR,
            "order": "000000001",
            "amount": D("10.56489").quantize(D(".01"), ROUND_HALF_UP),
            "merchant_data": "test merchant data",
            "merchant_name": "Example Commerce",
            "titular": "Example Ltd.",
            "product_description": "Products of Example Commerce",
            "merchant_url": "https://example.com/redsys/response",
        }
        self.request = Request(self.parameters)
        self.merchant_params = self.request.prepare_parameters()

    def test_encode_parameters(self):
        encoded_params = self.client.encode_parameters(self.merchant_params)
        encoded_params_str = encoded_params.decode("UTF-8")
        assert encoded_params_str.endswith("ZWRzeXMvcmVzcG9uc2UifQ==")
        assert len(encoded_params_str) == 624

    def test_decode_parameters(self):
        encoded_params = self.client.encode_parameters(self.merchant_params)
        decoded_params = self.client.decode_parameters(encoded_params)
        assert decoded_params == self.merchant_params

    def test_encrypt_3DES(self):
        encrypted_order = self.client.encrypt_3DES(self.request.order)
        assert len(encrypted_order) == 16
        assert (
            str(encrypted_order)
            == 'b"/\\x17\\\\6\\n\\x1aY\\xf3\\xb7\\\\\\xde\'d\\xa5\\x03\\xb7"'
        )

    def test_sign_hmac256(self):
        encrypted_order = self.client.encrypt_3DES(self.request.order)
        encoded_params = self.client.encode_parameters(self.merchant_params)
        signature = self.client.sign_hmac256(encrypted_order, encoded_params)
        assert signature == b"eEW4YN7/kkNHRO0/HhXy9TppzHwF38+eZApKAagDI9A="

    def test_prepare_request(self):
        prepared_request = self.client.prepare_request(self.parameters)
        assert len(prepared_request.keys()) == 3
        assert (
            prepared_request.get("Ds_Signature")
            == b"eEW4YN7/kkNHRO0/HhXy9TppzHwF38+eZApKAagDI9A="
        )
        pass

    def test_create_response_valid(self):
        merchant_parameters = {
            "Ds_MerchantCode": "100000001",
            "Ds_Terminal": "1",
            "Ds_TransactionType": "0",
            "Ds_Currency": 978,
            "Ds_Order": "000000001",
            "Ds_Amount": 1056,
            "Ds_MerchantData": "test merchant data",
        }
        encoded_params = self.client.encode_parameters(merchant_parameters)
        signature = self.client.generate_signature(self.request.order, encoded_params)
        response = self.client.create_response(
            signature.decode(), encoded_params.decode()
        )
        assert response._parameters.get("order") == merchant_parameters.get("Ds_Order")

    def test_create_response_invalid(self):
        merchant_parameters = {
            "Ds_MerchantCode": "100000001",
            "Ds_Terminal": "1",
            "Ds_TransactionType": "0",
            "Ds_Currency": 978,
            "Ds_Order": "000000001",
            "Ds_Amount": 1056,
            "Ds_MerchantData": "test merchant data",
        }
        encoded_params = self.client.encode_parameters(merchant_parameters)
        bad_signature = b"sadfe2r3q2fdssaf3"
        with pytest.raises(ValueError, match="The provided signature is not valid"):
            self.client.create_response(bad_signature.decode(), encoded_params.decode())



from decimal import ROUND_HALF_UP
from decimal import Decimal as D

import pytest

from redsys.constants import EUR, STANDARD_PAYMENT
from redsys.request import Request


class TestRequest:
    def test_create_request(self):
        parameters = {
            "merchant_code": "100000001",
            "terminal": "1",
            "transaction_type": STANDARD_PAYMENT,
            "currency": EUR,
            "order": "000000001",
            "amount": D("10.56489").quantize(D(".01"), ROUND_HALF_UP),
            "merchant_data": "test merchant data",
            "merchant_name": "Example Commerce",
            "titular": "Example Ltd.",
            "product_description": "Products of Example Commerce",
            "merchant_url": "https://example.com/redsys/response",
        }
        self.request = Request(parameters)
        assert self.request.product_description == "Products of Example Commerce"

    def test_bad_request_parameters(self):
        parameters = {
            "merchant_code": "100000001",
            "terminal": "1",
            "bad_key": "test",
        }
        with pytest.raises(ValueError, match="Unknown parameter"):
            self.request = Request(parameters)

    def test_prepare_parameters(self):
        parameters = {
            "merchant_code": "100000001",
            "terminal": "1",
            "transaction_type": STANDARD_PAYMENT,
            "currency": EUR,
            "order": "000000001",
            "amount": D("10.56489").quantize(D(".01"), ROUND_HALF_UP),
            "merchant_data": "test merchant data",
            "merchant_name": "Example Commerce",
            "titular": "Example Ltd.",
            "product_description": "Products of Example Commerce",
            "merchant_url": "https://example.com/redsys/response",
        }
        prepared_parameters = Request(parameters).prepare_parameters()
        assert len(prepared_parameters.keys()) == 11
        assert (
            prepared_parameters.get("Ds_Merchant_MerchantData") == "test merchant data"
        )



import base64
import hashlib
import hmac
import json
import re
from abc import ABC, abstractmethod
from typing import Any, Dict

from Crypto.Cipher import DES3

from redsys.request import Request
from redsys.response import Response

# Calculated parameters
SIGNATURE_VERSION = "Ds_SignatureVersion"
MERCHANT_PARAMETERS = "Ds_MerchantParameters"
SIGNATURE = "Ds_Signature"
DEFAULT_SIGNATURE_VERSION = "HMAC_SHA256_V1"


class Client(ABC):
    """
    Abstract class from which RedirectClient inherits.
    It implements methods that may be used in future clients(i.e. rest).
    """

    def __init__(self, secret_key: str = None):
        self.secret_key = secret_key

    @abstractmethod
    def create_response(
        self, signature, parameters, signature_version=DEFAULT_SIGNATURE_VERSION
    ):
        raise NotImplementedError

    @abstractmethod
    def prepare_request(self, request):
        raise NotImplementedError

    @staticmethod
    def encode_parameters(parameters: Dict[str, Any]) -> bytes:
        """Encodes the merchant parameters in base64"""
        return base64.b64encode(json.dumps(parameters).encode())

    @staticmethod
    def decode_parameters(parameters: bytes) -> Dict[str, Any]:
        """Decodes the merchant parameters from base64"""
        return json.loads(base64.b64decode(parameters).decode())

    @staticmethod
    def sign_hmac256(encrypted_order: bytes, merchant_parameters: bytes) -> bytes:
        """
        Generates the encrypted signature using the 3DES-encrypted order
        and base64-encoded merchant parameters
        """
        signature = hmac.new(
            encrypted_order, merchant_parameters, hashlib.sha256
        ).digest()
        return base64.b64encode(signature)

    def encrypt_3DES(self, order: str) -> bytes:
        """Encrypts(3DES algorithm) the payment order using the secret key"""
        cipher = DES3.new(
            base64.b64decode(self.secret_key), DES3.MODE_CBC, IV=b"\0\0\0\0\0\0\0\0"
        )
        # the cipher needs to be passed 16 bytes,
        # so "order" must be 16 bytes long,
        # therefore we left-justify adding ceros
        return cipher.encrypt(order.encode().ljust(16, b"\0"))

    def generate_signature(self, order: str, merchant_parameters: bytes) -> bytes:
        return self.sign_hmac256(self.encrypt_3DES(order), merchant_parameters)


class RedirectClient(Client):
    def prepare_request(self, parameters: Dict[str, Any]) -> Dict[str, Any]:
        """Takes the merchant parameters and returns the necessary parameters
        to make the POST request to Redsys"""
        request = Request(parameters)
        merchant_parameters = self.encode_parameters(request.prepare_parameters())
        signature = self.generate_signature(request.order, merchant_parameters)
        return {
            SIGNATURE_VERSION: DEFAULT_SIGNATURE_VERSION,
            MERCHANT_PARAMETERS: merchant_parameters,
            SIGNATURE: signature,
        }

    def create_response(
        self,
        signature: str,
        merchant_parameters: str,
    ) -> Response:
        """
        Decodes the Redsys response to check for validity.
        Checks if the received signature corresponds to the sent signature.

        Both the `signature` and `merchant parameters` are plain strings, not bytes.
        """
        decoded_parameters = self.decode_parameters(merchant_parameters)
        response = Response(decoded_parameters)
        calculated_signature = self.generate_signature(
            response.order, merchant_parameters.encode()
        )
        # Remove any non-alphanumeric characters from the signature
        not_alphanumeric = re.compile("[^a-zA-Z0-9]")
        safe_signature = re.sub(not_alphanumeric, "", signature)

        safe_calculated_signature = re.sub(
            not_alphanumeric, "", calculated_signature.decode()
        )
        if safe_signature != safe_calculated_signature:
            raise ValueError("The provided signature is not valid.")
        return response
